/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// rbp.h 

#ifndef rbp_h
#define rbp_h

#include <bp/bp.h>

class RBpConSpec;
class RBpUnit;
class RBpUnitSpec;

class RBpTrial;
class RBpSequence;

class APBpCycle;
class APBpSettle;
class APBpTrial;
class APBpMaxDa_De;

// _

// the unit maintains activation and other state information automatically
// and the size of this list is used to determine when to backpropagate
// depending on the parameters in the unit spec (time window, bp window).
// also, the process can excplicitly backprop at any given point

// instead of keeping big lists of all sorts of things, only the essential forward
// pass values are maintained, and the rest is propagated sequentially during the
// backpropagation phase
// thus, prv_xxx values are kept of all variables, and arrays of only the activations
// are kept.  at each act step, the time is shifted forward (cur becomes prv)
// and at each bp step, time is shifted backwards (prv becomes cur)

// forward passes: do Compute_Net() on whole net, then Compute_Act()
// store resulting act values, etc.

// backward passes: do Compute_Error, then Compute_Send_dEdNet, the Step_Back
// and iterate..

class RBpConSpec : public BpConSpec {
  // Recurrent Backprop Con Spec
public:
  inline void 		C_Compute_dWt(BpCon* cn, RBpUnit* ru, RBpUnit* su);
  inline void 		Compute_dWt(Con_Group* cg, Unit* ru);
  // Compute dE with respect to the weights (using prv_act) as sender

  void 	Initialize()		{ };
  void	Destroy()		{ };
  TA_BASEFUNS(RBpConSpec);
};


class RBpUnitSpec : public BpUnitSpec {
  // Recurrent Backprop Unit Specification
public:
  enum TimeAvgType {		// type of time-averaging to perform
    ACTIVATION,		// time-average the activations
    NET_INPUT 			// time-average the net inputs
  };

  float		dt;		// grain of time computing on (must be in [0..1] range)
  TimeAvgType	time_avg;	// type of time-averaging to perform
  bool		soft_clamp;
  float		soft_clamp_gain; // gain on the soft clamping
  bool		teacher_force;	 // use teacher forcing
  bool		store_states;
  // store activity states (usually true, except in AP mode)
  Random	initial_act;	 // initial activation value

  virtual void  Compute_ClampExt(RBpUnit* u);
  // compute activations resulting from clamped external input (for initial state)

  void		InitState(Unit* u);
  void		Compute_Net(Unit* u);
  virtual void	Compute_Act_impl(RBpUnit* u);
  void		Compute_Act(Unit* u);
  void		Compute_dEdA(BpUnit* u);
  void 		Compute_dEdNet(BpUnit* u);
  void 		Compute_dWt(Unit* u);
  void 		UpdateWeights(Unit* u);

  void	UpdateAfterEdit();
  void 	Initialize();
  void	Destroy()		{ };
  void	InitLinks();
  SIMPLE_COPY(RBpUnitSpec);
  COPY_FUNS(RBpUnitSpec, UnitSpec);
  TA_BASEFUNS(RBpUnitSpec);
};

class float_CircBuffer : public float_Array {
 // Circular buffer for holding state information
public:
  int		st_idx;		// starting index
  int		length;		// logical length of the list

  int		CircIdx(uint idx) const
  { int rval = idx+st_idx; if(rval >= size) rval -= size; return rval; }
  // gets real index based on circular buffer given logical index 

  void		ShiftLeft(int nshift) 
  { st_idx = CircIdx(nshift); length -= nshift; }

  void		Add(const float& item);
  // adds in the framework of a circular buffer

  float&	Peek() const
  { float* rval = (float*)&err; if(length > 0) rval = &(El(CircIdx(length-1)));
    return *rval; }

  void		Reset();

  void 	Initialize();
  void	Destroy()		{ };
  void 	Copy_(const float_CircBuffer& cp);
  COPY_FUNS(float_CircBuffer, float_Array);
  TA_BASEFUNS(float_CircBuffer);
};


class RBpUnit : public BpUnit {
  // recurrent BP unit
public:
  float		da;		// delta-activation (change in activation value)
  float		ddE;		// delta-delta-Error (change in error derivative)

  ExtType	prv_ext_flag;	// #HIDDEN previous external input flag
  float		prv_targ;	// #NO_VIEW #HIDDEN previous target value
  float		prv_ext;	// #NO_VIEW #HIDDEN previous external input value
  float		prv_act;	// #NO_VIEW #HIDDEN previous activation value
  float		act_raw;	// #NO_VIEW current raw activation value
  float		prv_net;	// #NO_VIEW #HIDDEN previous net-input value
  float		prv_dEdA;	// #NO_VIEW #HIDDEN previous dEdA value
  float		prv_dEdNet;	// #NO_VIEW #HIDDEN previous dEdNet value

  float_CircBuffer ext_flags;	// #NO_VIEW array of external_flag values
  float_CircBuffer targs;	// #NO_VIEW array of target values
  float_CircBuffer exts;	// #NO_VIEW array of external input values
  float_CircBuffer acts;	// #NO_VIEW array of activation values

  void 		InitExterns();	// keep prv_values..
  virtual void	StoreState();
  // store state information in buffers
  void		InitForBP()	{ prv_dEdA = prv_dEdNet = 0.0f; }
  // initialization prior to backpropagation (prv_dEdA, prv_dEdNet)
  virtual bool	StepBack(int tick);
  // move units backwards in time at given tick position in array
  virtual bool	RestoreState(int tick);
  // restore state of units to current values at given tick
  virtual void	ShiftBuffers(int ticks);
  // shift unit data buffers by given number of ticks

  // spec functions
  void    Compute_ClampExt() { ((RBpUnitSpec*)spec.spec)->Compute_ClampExt(this); }

  void 	Initialize();
  void	Destroy()		{ };
  void	InitLinks();
  void	Copy_(const RBpUnit& cp);
  COPY_FUNS(RBpUnit, BpUnit);
  TA_BASEFUNS(RBpUnit);
};

// use previous activation value
inline void RBpConSpec::C_Compute_dWt(BpCon* cn, RBpUnit* ru, RBpUnit* su) {
  cn->dEdW += su->prv_act * ru->dEdNet;
}
inline void RBpConSpec::Compute_dWt(Con_Group* cg, Unit* ru) {
  CON_GROUP_LOOP(cg,C_Compute_dWt((BpCon*)cg->Cn(i), (RBpUnit*)ru, (RBpUnit*)cg->Un(i)));
}

// perform a bp whenever the units have time_window amount of information in
// their buffers, then, shift left those buffers by bp_gap amount
// then, they'll be ready to go again when buffers have time_window in them again..
// this makes it so that the units keep track of themselves, and the process
// will interface nicely without requiring firm alignment between sequence periods
// and various time windows..
// indeed, one can run just the RBpTrial with a flat environment and it will work
// just fine..

// time gets set/reset by checking the buffers on the units: when they are 0
// at start of Compute_Act then time gets reset to 0.  otherwise, time increases
// by dt each step of processing (thus, it is not tied to a specific sequence length)

class RBpTrial : public BpTrial {
  // one presentation of an event to RBp
public:
  float		time;		
  // #READ_ONLY #SHOW current time (relative to start of sequence)
  float		dt;		
  // #READ_ONLY #SHOW this is made to correspond to the dt used by units
  float		time_window;	// time window to pay attention to derivatives for
  float		bp_gap;		// time period to go before performing a bp
  bool		real_time;
  // use 'real time' model (else time_window = length of sequence)
  bool		bp_performed;	// #READ_ONLY true if bp was just performed last step

  int		time_win_ticks;	// #READ_ONLY time window in ticks
  int		bp_gap_ticks;	// #READ_ONLY bp window in ticks

  void		Init_impl();
  void		Loop();
  bool		Crit()		{ return true; }
  
  virtual void  Compute_ClampExt();
  // compute clamped external activations (for zero time step)

  void		Compute_Act();
  void		Compute_dEdA_dEdNet();

  virtual void	StoreState();
  // store current state of network in buffers
  virtual void 	InitForBP();
  // initialize prior to performing BP
  virtual void	StepBack(int tick);
  // step back during BP process at given tick
  virtual void	RestoreState(int tick);
  // restore state to values at given point in time
  virtual void	ShiftBuffers();
  // shift buffers back by bp_gap_ticks
  virtual bool	PerformBP();
  // checks if its time to perform bp, does so, returns if it did or not
  virtual void	PerformBP_impl();
  // actually do the bp
  virtual int	GetUnitBufSize(bool in_updt_after = false);
  // finds first unit in first layer and gets current buffer size of that unit

  void		GetCntrDataItems();
  void		GenCntrLog(LogData* ld, bool gen);
  // output time information

  bool		CheckUnit(Unit* ck); // check for store_states too

  void	UpdateAfterEdit();
  void 	Initialize();
  void	InitLinks();
  void	Destroy()		{ };
  SIMPLE_COPY(RBpTrial);
  COPY_FUNS(RBpTrial, BpTrial);
  TA_BASEFUNS(RBpTrial);
};

class RBpSequence : public SequenceProcess {
  // one sequence of events, handles TimeEvents properly
public:
  virtual int	GetMaxTick();	// get maximum tick val based on current group

  void		Init_impl();
  void		GetEventList();
  void		GetCurEvent();
  void		Loop();

  void 	Initialize();
  void	Destroy()		{ };
  TA_BASEFUNS(RBpSequence);
};


class RBpSE_Stat : public SE_Stat {
  // Squared error for recurrent backprop, mulitplies by dt
public:
  void		Network_Stat();
  
  void	Initialize();
  void	Destroy()		{ };
  TA_BASEFUNS(RBpSE_Stat);
};


//////////////////////////////////////////
//	Almeida-Pineda Algorithm	//
//////////////////////////////////////////

class APBpCycle : public CycleProcess {
  // one cycle of processing in almeida-pineda (either act or bp depending on 'phase')
public:
  APBpSettle*	apbp_settle;
  // #NO_SUBTYPE #READ_ONLY #NO_SAVE pointer to parent settle proc
  APBpTrial* 	apbp_trial;
  // #NO_SUBTYPE #READ_ONLY #NO_SAVE pointer to parent phase trial

  void		Loop();
  bool		Crit()		{ return true; }
  
  virtual void	Compute_Act();	
  virtual void	Compute_Error();
  virtual void	Compute_dEdA_dEdNet();

  void 	UpdateAfterEdit();
  void 	Initialize();
  void 	Destroy()		{ CutLinks(); }
  void	CutLinks();
  TA_BASEFUNS(APBpCycle);
};

class APBpSettle : public SettleProcess {
  // one settling phase in Almeide-Pineda (either act or bp depending on phase)
public:
  APBpTrial* 	apbp_trial;
  // #NO_SUBTYPE #READ_ONLY #NO_SAVE pointer to parent phase trial

  void		Init_impl();	// initialize start of settling

  void 	UpdateAfterEdit();
  void 	Initialize();
  void 	Destroy()		{ CutLinks(); }
  void	InitLinks();
  void	CutLinks();
  SIMPLE_COPY(APBpSettle);
  COPY_FUNS(APBpSettle, SettleProcess);
  TA_BASEFUNS(APBpSettle);
};


class APBpTrial : public TrialProcess {
  // one Almeida-Pineda BP Trial
public:
  enum StateInit {		// ways of initializing the state of the network
    DO_NOTHING,			// do nothing
    INIT_STATE 			// initialize state
  };
    
  enum Phase {
    ACT_PHASE,			// activation phase
    BP_PHASE 			// backpropagation phase
  };
  
  Counter	phase_no;	// Current phase number
  Phase		phase;		// state variable for phase
  StateInit	trial_init;	// how to initialize network state at start of trial
  bool		no_bp_stats;	// don't do stats/logging in the bp phase
  bool		no_bp_test; 	// don't run the bp phase when testing


  virtual void	Compute_dWt();	// compute weight changes

  void		C_Code();	// modify to use the no_plus_stats flag
  
  void		Init_impl();
  void		UpdateState();
  bool		Crit()		{ return SchedProcess::Crit(); }
  void		Final();	// compute weight changes at end..
  
  void		GetCntrDataItems();
  void		GenCntrLog(LogData* ld, bool gen);

  bool		CheckUnit(Unit* ck); // check for store_states too

  void	Initialize();
  void	Destroy()		{ };
  void	InitLinks();
  void	Copy_(const APBpTrial& cp);
  COPY_FUNS(APBpTrial, TrialProcess);
  TA_BASEFUNS(APBpTrial);
};

class APBpMaxDa_De : public Stat {
  /* ##COMPUTE_IN_SettleProcess ##LOOP_STAT computes max of da and ddE to determine
     when to stop settling in almeida-pineda algorithm */
public:
  StatVal	da_de;		// max of delta-activation or delta-error

  void		RecvCon_Run(Unit*)	{ }; // don't do these!
  void		SendCon_Run(Unit*)	{ };

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Init();
  void 		Unit_Stat(Unit* unit);

  void 	Initialize();		// set minimums
  void	Destroy()		{ };
  SIMPLE_COPY(APBpMaxDa_De);
  COPY_FUNS(APBpMaxDa_De, Stat);
  TA_BASEFUNS(APBpMaxDa_De);
};

//////////////////////////////////////////
//	Additional Unit Types		//
//////////////////////////////////////////

class NoisyRBpUnitSpec : public RBpUnitSpec {
  // RBp with noisy output signal (act plus noise)
public:
  Random	noise;		// what kind of noise to add to activations
  float		sqrt_dt; 	// #HIDDEN square-root of dt for noise

  void 		Compute_Act_impl(RBpUnit* u);

  void	UpdateAfterEdit();
  void	Initialize();
  void 	Destroy()		{ };
  void 	InitLinks();
  SIMPLE_COPY(NoisyRBpUnitSpec);
  COPY_FUNS(NoisyRBpUnitSpec, RBpUnitSpec);
  TA_BASEFUNS(NoisyRBpUnitSpec);
};

#endif // rbp_h

