/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// pdplog.cc
#include <pdp/pdplog.h>
#include <pdp/process.h>
#include <pdp/sched_proc.h>
#include <ta/ta_group_iv.h>
#include <values.h>

#include <ta/enter_iv.h>
#include <IV-look/kit.h>
#include <InterViews/style.h>
#include <ta/leave_iv.h>


////////////////////////////////////////////////////////////////////////////////
// important notes on data_range, log_lines, and view_range:
// data_range reflects log_lines if logging to file, otherwise it is in lines
//   	in the data object
// data_range.min is >= 0, but data_range.max is not the _size_ of the max
// 	but the _line number_ of the max:  thus it is typically log_lines - 1
//	or data.MaxLength() - 1
// view_range is in the same units as data_range
// to translate either view or data_range into indicies in the data table (data)
// 	it is necessary to subtract data_range.min from the view_range
// 	to take into account the possiblity that the data in the table is only
// 	a portion of the data in the actual log file.
////////////////////////////////////////////////////////////////////////////////


//////////////////////////
// 	PDPLog		//
//////////////////////////
 
void PDPLog::Initialize() {
  log_file.cmp = false;
  log_file.mode = taFile::NO_AUTO;
  log_lines = 0;

  data.el_typ = &TA_float_Data;
  data_bufsz = 1000;		// default bufsz is 1000 lines
  data_shift = .20f;		// shift 20 percent (100 lines)
  record_proc_name = false;	// don't by default..
  data_range.min = 0;
  data_range.max = -1;
  cur_proc = NULL;
  views.SetBaseType(&TA_LogView);
}

void PDPLog::InitLinks() {
  taBase::Own(log_proc, this);
  taBase::Own(log_file, this);
  taBase::Own(data, this);
  taBase::Own(data_range, this);
  taBase::Own(log_data, this);

  WinMgr::InitLinks();
  InitFile();
  log_file.Close();
}

void PDPLog::Destroy() {
  CutLinks();
}

void PDPLog::CutLinks() {
  cur_proc = NULL;
  RemoveAllUpdaters();
  WinMgr::CutLinks();
}

void PDPLog::Copy_(const PDPLog& cp) {
  log_file = cp.log_file;
  log_lines = cp.log_lines;
  log_data = cp.log_data;

  data = cp.data; 
  data_bufsz = cp.data_bufsz;
  data_shift = cp.data_shift;
  data_range = cp.data_range;
  record_proc_name = cp.record_proc_name;

  log_proc.BorrowUnique(cp.log_proc);	// this is a link group..
  cur_proc = cp.cur_proc;
  InitAllViews();
}

void PDPLog::AddUpdater(SchedProcess* sp) {
  if(sp == NULL) return;
  bool was_linked = false;
  was_linked = sp->logs.LinkUnique(this);	// add this to the list on the process
  was_linked = (log_proc.LinkUnique(sp) || was_linked); // add the process to the list on this;
  if(was_linked) {
    name = sp->name + "_" + GetTypeDef()->name;
    SyncLogViewUpdaters();
    GetHeaders();
  }
}  

void PDPLog::RemoveUpdater(SchedProcess* sp) {
  if(sp == NULL) return;
  sp->logs.RemoveLeaf(this);	// remove this to the list on the process
  log_proc.RemoveLeaf(sp);     	// remove the process to the list on this;
  SyncLogViewUpdaters();
  GetHeaders();
}  

void PDPLog::RemoveAllUpdaters() {
  int i;
  for(i=0; i<log_proc.size; i++) {
    ((SchedProcess*)log_proc.FastEl(i))->logs.RemoveLeaf(this);
  }
  log_proc.RemoveAll();
  SyncLogViewUpdaters();
}

void PDPLog::SyncLogViewUpdaters() {
  LogView* lv;
  taLeafItr i;
  FOR_ITR_EL(LogView, lv, views., i) {
    lv->updaters.Reset();
    lv->updaters.BorrowUnique(log_proc);
  }
}

int PDPLog::Dump_Save_Value(ostream& strm, TAPtr par, int indent) {
  if(log_file.IsOpen())
    log_file.mode = taFile::APPEND; // auto-open in append mode..
  else
    log_file.mode = taFile::NO_AUTO;
  return WinMgr::Dump_Save_Value(strm, par, indent);
}

void PDPLog::UpdateAfterEdit() {
  WinMgr::UpdateAfterEdit();
  SyncLogViewUpdaters();
  LogView* lv = (LogView*)views[0];
  if(lv != NULL)
    lv->NotifyAllUpdaters();	// make sure logs are getting it from  us..
  if(log_proc.size > 1)		// always record proc name if more than one..
    record_proc_name = true;
  if(log_file.gf != NULL) {
    log_file.AutoOpen();
    log_file.mode = taFile::NO_AUTO; // prevent this from happening again..
  }
  if(log_file.IsOpen()) {
    if(log_lines == 0) {
      GetFileLength();
      if(log_file.gf->ostrm != NULL) {
	HeadToFile();
      }
    }
  }
  else {
    log_lines = 0;
    data_range.max = data.MaxLength()-1;
    data_range.min = 0;
  }
  if((log_proc.size > 0) && (name.empty() || name.contains(GetTypeDef()->name))) {
    name = ((SchedProcess*)log_proc[0])->name + "_" + GetTypeDef()->name;
  }
}

void PDPLog::InitFile() {
  if(log_file.gf == NULL) {
    if(!taMisc::iv_active)
      log_file.GetGetFile();
    else {
      log_file.gf = new taivGetFile(".","*log*",NULL,wkit->style(),false);
      taRefN::Ref(log_file.gf);
    }
  }
  log_file.fname = "";
  log_file.UpdateGF();
}

void PDPLog::SetSaveFile(const char* nm, bool no_dlg) {
  log_file.SaveAsFile(nm, no_dlg);
  GetFileLength();
  HeadToFile();
  if(taMisc::iv_active)
    SetWinName();
}

void PDPLog::SetAppendFile(const char* nm, bool no_dlg) {
  log_file.AppendFile(nm, no_dlg);
  GetFileLength();
  HeadToFile();
  if(taMisc::iv_active)
    SetWinName();
}

void PDPLog::LoadFile(const char* nm, bool no_dlg) {
  log_file.OpenFile(nm, no_dlg);
  GetFileLength();
  ReadNewLogFile();
  InitAllViews();
  if(taMisc::iv_active)
    SetWinName();
}

void PDPLog::CloseFile() {
  log_file.CloseFile();
  log_lines = 0;
  if(taMisc::iv_active)
    SetWinName();
}


//////////////////////////
//	Headers 	//
//////////////////////////

void PDPLog::NewHead(LogData& ld, SchedProcess* sproc) {
  // null is a sign to regenerate regardless
  if((sproc != NULL) && (cur_proc == sproc) && (ld.items.size == data.leaves))
    return; 

  if(sproc != NULL) {		// only send to log if its real
    cur_proc = sproc;
    HeadToLogFile(ld);
  }
  
  HeadToBuffer(ld);
  LogDataFromBuffer();		// get the log data corresponding to the buffer
  UpdateViewHeaders();		// tells views that header is updated
}

void PDPLog::UpdateViewHeaders() {
  LogView* lv;
  taLeafItr i;
  FOR_ITR_EL(LogView, lv, views., i) {
    lv->NewHead();
  }
}

void PDPLog::SetFieldHead
(LogData& ld, int fld, int top_cnt, int subgp_n, int subgp_max, int subgp_cnt) {
  DataTable* cdt = &data;
  if(subgp_max > 0) {
    if(data.gp.size <= subgp_n) {
      cdt = (DataTable*)data.NewGp(1);
    }
    else
      cdt = (DataTable*)data.gp[subgp_n];
    top_cnt = subgp_cnt;	// use top_cnt as index into cdt group
    while(cdt->size > subgp_max) { // keep it the same size..
      cdt->Remove(cdt->size-1);
    }
  }

  if(cdt->size <= top_cnt) {
    top_cnt = cdt->size;		// just in case
    if(ld.IsString(fld))		// indicates a string field
      cdt->NewEl(1,&TA_String_Data);
    else
      cdt->NewEl(1,&TA_float_Data);
  }
  if(ld.IsString(fld)) { // this checks to make sure the array is of the
			 // correct type
    if(!cdt->FastEl(top_cnt)->InheritsFrom(&TA_String_Data)) {
      DataArray_impl* nda = new String_Data;
      cdt->Replace(top_cnt, nda);
    }
  }
  else if(!cdt->FastEl(top_cnt)->InheritsFrom(&TA_float_Data)) {
    // this checks to make sure the array is of the correct type
    DataArray_impl* nda = new float_Data;
    cdt->Replace(top_cnt, nda);
  }

  DataArray_impl* da = cdt->FastEl(top_cnt);
  if(da->AR() == NULL) {
    da->NewAR();
    da->AR()->Alloc(data_bufsz);
  }
  da->name = ld.items.FastEl(fld)->name;
  da->disp_opts = ld.items.FastEl(fld)->disp_opts;
  if(cdt->name.empty()) // this gives a reasonable name to at least monitorstats
    cdt->name = da->name.after(".");
}

void PDPLog::HeadToBuffer(LogData& ld) {
  int subgp_n = -1;		// index of current sub-group
  int subgp_max = 0;		// max number in current sub-group
  int subgp_cnt = -1;		// number in current sub-group
  int top_cnt = 0;		// number in top-level group
  int i;
  for(i=0; i< ld.items.size; i++) {
    if(ld.IsVec(i)) {
      subgp_n++;
      subgp_max = ld.GetVecN(i);
      subgp_cnt = 0;
    }
    SetFieldHead(ld, i, top_cnt, subgp_n, subgp_max, subgp_cnt);
    if((subgp_max > 0) && (subgp_cnt < (subgp_max - 1)))
      subgp_cnt++;
    else {
      if(subgp_max > 0)
	subgp_max = 0;
      else
	top_cnt++;		// if just a gp, not a new item..
    }
  }
  while(data.size > top_cnt) { // keep it the same size..
    data.Remove(data.size-1);
  }
}

void PDPLog::HeadToLogFile(LogData& ld) {
  if(!log_file.IsOpen() || (log_file.gf->ostrm == NULL))
    return;
  
  ostream& strm = *(log_file.gf->ostrm);

  if(record_proc_name && (cur_proc != NULL))
    strm << cur_proc->name;
  strm << "_H:\t";
  int i;
  for(i=0; i < ld.items.size; i++) {
    DataItem* it = ld.items.FastEl(i);
    if(it->HasDispOption(" NARROW, "))
      LogColumn(strm, it->name, 1); // indicates one column
    else
      LogColumn(strm, it->name, 2); // default is 2 columns
  }
  strm << "\n";
  strm.flush();
}

DataItem* PDPLog::DataItemFromDataArray(DataArray_impl* da) {
  DataItem* it = new DataItem();
  it->name = da->name;
  it->disp_opts = da->disp_opts;
  if(da->InheritsFrom(TA_String_Data))
    it->is_string = true;
  log_data.items.Add(it);
  return it;
}

void PDPLog::LogDataFromDataTable(DataTable* dt, int st_idx) {
  int i;
  for(i=st_idx; i < dt->size; i++) {
    DataArray_impl* da = dt->FastEl(i);
    DataItemFromDataArray(da);
  }
  if(dt->gp.size > 0) {
    for(i=0; i<dt->gp.size; i++) {
      DataTable* ndt = (DataTable*)dt->Gp(i);
      if(ndt->size > 0) {
	DataArray_impl* da = ndt->FastEl(0);
	DataItem* it = DataItemFromDataArray(da);
	it->vec_n = ndt->size;
	LogDataFromDataTable(ndt, 1); // start from 1 since 0 was already done..
      }
    }
  }
}

void PDPLog::LogDataFromBuffer() {
  log_data.Reset();
  LogDataFromDataTable(&data);
  log_data.InitBlankData();
}

//////////////////////////
//	  Data 		//
//////////////////////////


void PDPLog::NewData(LogData& ld, SchedProcess* sproc) {
  NewHead(ld, sproc);		// see if header info is new..

  // buffer always reflects latest contents (views can scroll within)
  if(log_file.IsOpen() && (data_range.max < log_lines -1))
    Buffer_FF();

  DataToBuffer(ld);

  if(sproc != NULL)
    DataToLogFile(ld);

  int max_lns = data.MaxLength() -1; // get new size
  if(max_lns < data_range.Range()) { // must have shifted buffers..
    if(!log_file.IsOpen()) {
      data_range.min = 0;
      data_range.max = max_lns;
    }
    else {
      data_range.max++;
      data_range.min = data_range.max - max_lns;
      data_range.MinGT(0);
    }
    if(taMisc::iv_active) {
      LogView* lv;
      taLeafItr i;
      FOR_ITR_EL(LogView, lv, views., i) {
	lv->View_FF();		// catch the views up with the new range
      }
    }
  }
  else {
    data_range.max++;
    LogView* lv;
    taLeafItr i;
    FOR_ITR_EL(LogView, lv, views., i) {
      lv->NewData();		// just update new line
    }
  }
}


void PDPLog::SetFieldData
(LogData& ld, int fld, int top_cnt, int subgp_n, int subgp_max, int subgp_cnt) {
  DataTable* cdt = &data;
  if(subgp_max > 0) {
    if(data.gp.size <= subgp_n)	return;
    cdt = (DataTable*)data.gp[subgp_n];
    top_cnt = subgp_cnt;	// use top_cnt as index into cdt group
  }

  if(cdt->size <= top_cnt) return;

  DataArray_impl* da = cdt->FastEl(top_cnt);
  if(da->AR() == NULL)
    da->NewAR();
  if(da->AR()->size >= data_bufsz) { // adding will 'overflow' buffer
    da->AR()->ShiftLeftPct(data_shift);
  }

  if(ld.IsString(fld)) {
    if(da->InheritsFrom(&TA_String_Data))
      ((String_Array*)da->AR())->Add(ld.GetString(fld));
  }
  else {
    if(da->InheritsFrom(&TA_float_Data))
      ((float_Array*)da->AR())->Add(ld.GetFloat(fld));
  }
}

void PDPLog::DataToBuffer(LogData& ld) {
  int subgp_n = -1;		// index of current sub-group
  int subgp_max = 0;		// max number in current sub-group
  int subgp_cnt = -1;		// number in current sub-group
  int top_cnt = 0;		// number in top-level group
  int i;
  for(i=0; i< ld.items.size; i++) {
    if(ld.IsVec(i)) {
      subgp_n++;
      subgp_max = ld.GetVecN(i);
      subgp_cnt = 0;
    }
    SetFieldData(ld, i, top_cnt, subgp_n, subgp_max, subgp_cnt);
    if((subgp_max > 0) && (subgp_cnt < (subgp_max - 1)))
      subgp_cnt++;
    else {
      if(subgp_max > 0)
	subgp_max = 0;
      else
	top_cnt++;		// if just a gp, not a new item..
    }
  }
}

void PDPLog::DataToLogFile(LogData&) {
  if(!log_file.IsOpen() || (log_file.gf->ostrm == NULL))
    return;

  ostream& strm = *(log_file.gf->ostrm);

  if(record_proc_name && (cur_proc != NULL))
    strm << cur_proc->name;
  strm << "_D:\t";

  taLeafItr i;
  DataArray_impl* da;
  FOR_ITR_EL(DataArray_impl, da, data., i) {
    String el;
    taArray_base* ar = da->AR();
    if((ar != NULL) && (ar->size > 0))
      el = ar->El_GetStr_(ar->FastEl_(ar->size-1));
    else
      el = "n/a";

    if(da->HasDispOption(" NARROW, "))
      LogColumn(strm, el, 1);
    else
      LogColumn(strm, el, 2);
  }
  strm << "\n";
  strm.flush();
  log_lines++;
}

void PDPLog::GetHeaders() {
  data.Reset();		// clear out old data...

  LogView* lv;
  taLeafItr vi;
  FOR_ITR_EL(LogView, lv, views., vi) {
    if(lv->viewspec != NULL)
      lv->viewspec->Reset();	// clear all viewspecs too
    lv->view_range.min =0; lv->view_range.max = -1;
    lv->InitDisplay();
  }

  taLeafItr i;
  SchedProcess* sproc;
  FOR_ITR_EL(SchedProcess, sproc, log_proc., i) {
    sproc->GenLogData();
    NewHead(sproc->log_data, NULL); // this is a "fake" head, not from a log directly
    HeadToFile();
  }
}

ostream& PDPLog::LogColumn(ostream& strm, String& str, int tabs) {
  strm << str;
  if(tabs == 2) {		// only 2 options (1 or 2)
    if(str.length() < 8)
      strm << "\t\t";
    else if(str.length() < 16)
      strm << "\t";
    else
      strm << "\t";		// always tab delimited..
  }
  else {
    if(str.length() < 8)
      strm << "\t";
    else
      strm << "\t";
  }
  return strm;
}
  
 

//////////////////////////
// 	Buffers		//
//////////////////////////
 
// checks for comments and headers,
// 0 = comment, 1 = header, 2 = data

int PDPLog::LogLineType(char* lnc) {
  String ln = String(lnc);

  // skip leading spaces if any
  while(ln.firstchar() == ' ') {
    ln = ln.after(' ');
  }
  if(ln.empty())  return 0;

  if((ln.firstchar() == '#') || (ln.before(2) == "//")) {
    return 0;
  }
  
  String id = ln.before('\t');
  if(id .empty()) return 0;

  String id_sufx = id.after('_', -1); // find after last underbar
  if(id_sufx == "H:") return 1;
  if(id_sufx == "D:") return 2;
  return 2;			// assume its data
}

// reads the file up to just before the dataline, strm must be just opened
// returns the maximum dataline number it could read

int PDPLog::FileScanTo(istream& strm, int log_ln) {
  if(log_ln <= 0) return 0;	// ready to read line 0

  int count = 0;
  String templine;
  
  while(readline(strm,templine)) {
    if(LogLineType(templine) != 2) // only count data lines
      continue;
    count++;
    if (count >= log_ln) break;
  }
  return count;		// the number of the line were ready to read
}  


int PDPLog::GetFileLength() {
  if(!log_file.IsOpen()) return 0;

  taivGetFile gf;		// use a getfile for compressed reads..
  gf.fname = log_file.fname;
  istream* strm = gf.open_read();
  if((strm == NULL) || strm->bad())
    return 0;

  int act_lines = FileScanTo(*strm, MAXINT); // read as far as we can..
  gf.Close();

  log_lines = act_lines;
  data_range.MaxLT(log_lines -1);
  return act_lines;
}

void PDPLog::HeadFromLogFile(String& hdln) { // argument is mutable..
  int itm_cnt = 0;
  if(hdln.lastchar() != '\t') hdln += "\t";
  while(hdln.contains('\t')) {
    String f = hdln.before('\t');
    if((!f.empty()) && !f.contains("_H:")) {
      DataItem* it;
      if(itm_cnt >= log_data.items.size) {
	it = new DataItem();
	log_data.items.Add(it);
      }
      else
	it = log_data.items.FastEl(itm_cnt);
      if(f != it->name) {	// names don't match
	it->Initialize();	// reset info on 
	it->name = f;
	if(f[0] == '$') {
	  it->is_string = true;
	  if(f[1] == '<') {
	    String vec_n = f.after('<');
	    vec_n = vec_n.before('>');
	    it->vec_n = (int)vec_n;
	  }
	}
	else if(f[0] == '|') {
	  it->AddDispOption("NARROW, ");
	}
	else if(f[0] == '<') {
	  String vec_n = f.after('<');
	  vec_n = vec_n.before('>');
	  it->vec_n = (int)vec_n;
	}
      }
      itm_cnt++;
    }
    hdln=hdln.after('\t');
  }
  if(log_data.index.size != log_data.items.size)
    log_data.InitBlankData();
  NewHead(log_data, NULL);
}


void PDPLog::DataFromLogFile(String& dtln) {
  // log data heads are assumed to be current..
  int cnt = 0;
  if(dtln.lastchar() != '\t') dtln += "\t";
  if(log_data.index.size != log_data.items.size) { // make sure these line up..
    log_data.InitBlankData();			   // if not, reset data
  }
  while(dtln.contains('\t') && (cnt < log_data.items.size)) {
    String f = dtln.before('\t');
    if((!f.empty()) && !f.contains("_D:")) {
      if(log_data.IsString(cnt))
	log_data.GetString(cnt) = f; // these don't existe..
      else
	log_data.GetFloat(cnt) = atof(f);
      cnt++;
    }
    dtln=dtln.after('\t');
  }
  DataToBuffer(log_data);
}


void PDPLog::LogFileToBuffer() {
  if(!log_file.IsOpen()) return;

  data.ResetData();		// clear out the arrays for new data
  data_range.MinGT(0);		// min must be greater than 0

  taivGetFile gf;		// use a getfile for compressed reads..
  gf.fname = log_file.fname;
  istream* strm = gf.open_read();

  if((strm == NULL) || strm->bad())
    return;

  int act_lines = FileScanTo(*strm, (int)data_range.min);
  if(act_lines != data_range.min) {
    log_lines = act_lines;	// we adjust to the actual size of the file..
    data_range.MaxLT(log_lines-1);
    data_range.min = data_range.max - data_bufsz + 1;
    data_range.MinGT(0);
    if(log_lines == 0) {
      gf.Close();
      return;
    }
  }

  int lno = data_range.min;
  String templn;
  int rval;
  while((rval = readline(*strm,templn)) && (lno <= data_range.max)) {
    int ltyp = LogLineType(templn);
    if(ltyp == 0)
      continue;
    if(ltyp == 1) {
      HeadFromLogFile(templn);
    }
    if(ltyp == 2) {
      DataFromLogFile(templn);
      lno++;
    }
  }
  if(rval == 0) {
    log_lines = lno;
    data_range.MaxLT(lno -1 );	// reset max based on what we actually got
  }
  gf.Close();
}

void PDPLog::HeadToFile() {	// based on buffer
  if(!log_file.IsOpen() || (log_file.gf->ostrm == NULL))
    return;

  ostream& strm = *(log_file.gf->ostrm);

  if(record_proc_name && (cur_proc != NULL))
    strm << cur_proc->name;
  strm << "_H:\t";
  int i;
  for(i=0; i < data.leaves; i++) {
    DataArray_impl* da = data.Leaf(i);
    int wdth = 2;
    if(da->HasDispOption(" NARROW, "))
      wdth = 1;
    LogColumn(strm, da->name, wdth);
  }
  strm << "\n";
  strm.flush();
}

void PDPLog::BufferToFile() {
  if(!log_file.IsOpen() || (log_file.gf->ostrm == NULL))
    return;

  ostream& strm = *(log_file.gf->ostrm);

  int max_lns = data.MaxLength()-1;
  int ln;
  for(ln=0; ln <= max_lns; ln++) {
    if(record_proc_name && (cur_proc != NULL))
      strm << cur_proc->name;
    strm << "_D:\t";
    int from_end = max_lns - ln ;
    taLeafItr i;
    DataArray_impl* da;
    FOR_ITR_EL(DataArray_impl, da, data., i) {
      String el;
      taArray_base* ar = da->AR();
      if((ar != NULL) && (ar->size > from_end) && (from_end >= 0))
	el = ar->El_GetStr_(ar->FastEl_(ar->size-1-from_end));
      else
	el = "n/a";

      if(da->HasDispOption(" NARROW, "))
	LogColumn(strm, el, 1);
      else
	LogColumn(strm, el, 2);
    }
    strm << "\n";
  }
  strm.flush();
}

void PDPLog::Buffer_F() {
  if(!log_file.IsOpen()) return;
  int shft = (int)((float)data_bufsz * data_shift);
  data_range.max += shft;
  data_range.MaxLT(log_lines - 1);
  data_range.min = data_range.max - data_bufsz + 1;
  data_range.MinGT(0);
  LogFileToBuffer();
}

void PDPLog::Buffer_FF() {
  if(!log_file.IsOpen()) return;
  data_range.max = log_lines - 1;
  data_range.min = data_range.max - data_bufsz + 1;
  data_range.MinGT(0);
  LogFileToBuffer();
}

void PDPLog::Buffer_R() {
  if(!log_file.IsOpen()) return;
  int shft = (int)((float)data_bufsz * data_shift);
  data_range.min -= shft;
  data_range.MinGT(0);
  data_range.max = data_range.min + data_bufsz - 1;
  data_range.MaxLT(log_lines - 1);
  LogFileToBuffer();
}

void PDPLog::Buffer_FR() {
  if(!log_file.IsOpen()) return;
  data_range.min = 0;
  data_range.max = data_range.min + data_bufsz -1;
  data_range.MaxLT(log_lines - 1);
  LogFileToBuffer();
}

void PDPLog::ReadNewLogFile() {
  Buffer_FF();		// go to end
  if(!taMisc::iv_active) return;
  LogView* lv;
  taLeafItr i;
  FOR_ITR_EL(LogView, lv, views., i) {
    lv->View_FF();		// same for all the views
  }
}

void PDPLog::Clear() {
  data.ResetData();
  data_range.min = 0; data_range.max = -1;
  if(!taMisc::iv_active) return;
  LogView* lv;
  taLeafItr i;
  FOR_ITR_EL(LogView, lv, views., i)
    lv->Clear_impl();
}

void NetLog::SetNetwork(Network* net) {
  NetLogView* lv;
  taLeafItr i;
  FOR_ITR_EL(NetLogView, lv, views., i)
    lv->SetNetwork(net);
}


