/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// procs_extra.h
// extra process (stats, schedule, etc) objects derived from basic ones that are
// included in standard pdp library

#ifndef procs_extra_h
#define procs_extra_h

#include <pdp/sched_proc.h>

//////////////////////////////////
//	Basic Processes		//
//////////////////////////////////

class ScriptProcess : public Process {
  // a process for use with scripts (has s_args)
public:
  String_Array	s_args;		// string-valued arguments to pass to script

  void	UpdateAfterEdit();
  void 	InitLinks();
  void	Copy_(const ScriptProcess& cp);
  COPY_FUNS(ScriptProcess, Process);
  TA_BASEFUNS(ScriptProcess);
};

class SaveNetsProc : public Process {
  // saves networks with network.name + counters (batch, epoch)
public:
  void			C_Code();
  
  void	Initialize()	{ };
  void	Destroy()	{ };
  TA_BASEFUNS(SaveNetsProc);
};

class SaveWtsProc : public Process {
  // saves weights with network.name + counters (batch, epoch)
public:
  void			C_Code();
  
  void	Initialize()	{ };
  void	Destroy()	{ };
  TA_BASEFUNS(SaveWtsProc);
};

class InitWtsProc : public Process {
  // initialize the network's weights (InitWtState)
public:
  void			C_Code();
  
  void	Initialize()	{ };
  void	Destroy()	{ };
  TA_BASEFUNS(InitWtsProc);
};


//////////////////////////////////
//	Schedule Processes	//
//////////////////////////////////

class SyncEpochProc : public EpochProcess {
  // Epoch that runs two different sub-processes
public:
  Network*		second_network;	  // #CONTROL_PANEL the network for the second trial process
  TypeDef*		second_proc_type; // #TYPE_SchedProcess type of second process
  SchedProcess*		second_proc; 	  // #CONTROL_PANEL second process to call

  void		Init();
  void		Init_flag();
  void		Init_force();
  void		Loop();		// do on-line, small_batch here
  void		Final();	// batch wt update here

  SchedProcess*	FindSubProc(TypeDef* td);
  void		CreateSubProcs(bool update);

  void	SetDefaultPNEPtrs();

  void 	Initialize();	
  void 	Destroy();
  void	InitLinks();
  void	CutLinks();
  void	Copy_(const SyncEpochProc& cp);
  COPY_FUNS(SyncEpochProc, EpochProcess);
  TA_BASEFUNS(SyncEpochProc);
};

class GridSearchBatch : public BatchProcess {
  // For searching parameter space over network trainings
public:
  float		start_val;	// #CONTROL_PANEL start value of parameter being searched
  float		inc_val;	// #CONTROL_PANEL increment of parameter being searched
  float		cur_val;	// #CONTROL_PANEL #IV_READ_ONLY current value of parameter based on batch value
  String	param_path;	// path to the parameter (starting at the project)
  
  virtual void	Compute_CurVal(); // compute the current value from batch counter
  virtual void	SetParamVal();	  // set parameter value to be current value

  void		Init_impl();
  void		Loop();

  void		GetCntrDataItems();
  void		GenCntrLog(LogData* ld, bool gen);

  void	Initialize();
  void	Destroy()		{ };
  SIMPLE_COPY(GridSearchBatch);
  COPY_FUNS(GridSearchBatch, BatchProcess);
  TA_BASEFUNS(GridSearchBatch);
};


//////////////////////////////////
//	Fork and Bridge		//
//////////////////////////////////

class ForkProcess : public SchedProcess {
  // Runs two different sub-processes
public:
  Network*		second_network;	  // the network for the second process
  Environment*		second_environment; // the environment for the second process
  TypeDef*		second_proc_type; // #TYPE_SchedProcess #NULL_OK type of second process
  SchedProcess*		second_proc; 	  // second process to call

  void		Init();
  void		Init_flag();
  void		Init_force();
  void		Loop();
  bool		Crit()	{ return true; }

  SchedProcess*	FindSubProc(TypeDef* td);
  void		CreateSubProcs(bool update);

  void	SetDefaultPNEPtrs();

  void 	Initialize();	
  void 	Destroy();
  void	InitLinks();
  void	CutLinks();
  void	Copy_(const ForkProcess& cp);
  COPY_FUNS(ForkProcess, SchedProcess);
  TA_BASEFUNS(ForkProcess);
};

class BridgeProcess : public Process {
  // a bridge that connects two different networks together by copying values
public:
  enum BridgeDirection {
    ONE_TO_TWO,			// first network (network) copies to second_network
    TWO_TO_ONE 			// second_network copies to first one (network)
  };

  Network*	second_network;	  // the other network to bridge to
  Layer*	src_layer;	  // #HIDDEN #NO_SAVE the actual source layer
  Layer*	trg_layer;	  // #HIDDEN #NO_SAVE the actual target layer
  BridgeDirection direction;	  // direction to copy in
  String	src_layer_nm;	  // name of the source layer
  String	trg_layer_nm;	  // name of the target layer
  String	src_variable;	  // the source variable (member) to copy
  String	trg_variable;	  // the target variable (member) to copy
  Unit::ExtType	trg_ext_flag;	  // flag to apply to the target units

  virtual bool	SetLayerPtrs();	// set the layer pointers from names, direction

  void		C_Code();
  
  void	UpdateAfterEdit();
  void 	Initialize();	
  void 	Destroy();
  void	InitLinks();
  void	CutLinks();
  void	Copy_(const BridgeProcess& cp);
  COPY_FUNS(BridgeProcess, Process);
  TA_BASEFUNS(BridgeProcess);
};

// Note: following two types contributed by Steve Pope @ the Prediction Company

class MultiEnvProcess : public SchedProcess {
  // SchedProcess which will run subprocess over multiple environments, indexed by counter
public:
  Environment_List	environments;   // #LINK_GROUP list of environments to process
  Counter		env;            // current environment number
  
  void 		Init_impl();
  void 		UpdateState();
  virtual void 	SetCurEnvironment();    // set up environment from the counter

  void UpdateAfterEdit();
  void Initialize();   
  void Destroy();
  void InitLinks();
  void CutLinks();
  void Copy_(const MultiEnvProcess& cp);
  COPY_FUNS(MultiEnvProcess, SchedProcess);
  TA_BASEFUNS(MultiEnvProcess);
};

class PatternFlagProcess : public SchedProcess {
  /* iteratively sets/resets pattern flag at index = counter for all patterns,
     e.g. to determine sensitivity to given input */
public:
  int          		pattern_no;     // index of pattern to update
  PatternSpec::PatFlags flag;   	// flag to be set (or unset if invert)
  bool			invert;		// unset the flag instead of setting it
  Counter		val_idx;	// index of current pattern value to be flagged
  
  void 		Init_impl();
  void 		Loop();
  virtual void 	SetCurFlags(int* old_flag);
  // set up flags for current counter (sets old_flag if provided to prev flag val)
  virtual void 	ClearCurFlags(int* old_flag);
  // clear flags for current counter (resetting to old_flag if provided)

  void UpdateAfterEdit();
  void Initialize();   
  void Destroy()	{ };
  void InitLinks();
  void Copy_(const PatternFlagProcess& cp);
  COPY_FUNS(PatternFlagProcess, SchedProcess);
  TA_BASEFUNS(PatternFlagProcess);
};

  
//////////////////////////////////
//	Stats Processes		//
//////////////////////////////////

class ClosestEventStat : public Stat {
  // ##COMPUTE_IN_TrialProcess gets the closest event to the current output activity pattern
public:
  TrialProcess* trial_proc;	// #READ_ONLY #NO_SAVE current trial process
  float_RArray::DistMetric cmp_type;	// comparison type to perform to compute distance
  float		dist_tol;	// tolerance value for distance functions
  bool		norm;		// whether to normalize (distance, inner prod)
  int		subgp_no;	// if not -1, then indicates which subgroup of units within layer
  StatVal	dist;		// distance from closest event
  StatVal	ev_nm;		// name of event which was closest
  StatVal	sm_nm;		// 1 or 0 depending on whether the name matched cur event
  float_RArray	lay_act;	// #HIDDEN #NO_SAVE buffer for layer activity

  void		InitStat();
  void		Init();
  bool		Crit();

  void		Layer_Run();
  void		Layer_Stat(Layer* lay);

  void		RecvCon_Run(Unit*)	{}; // don't do these!
  void		SendCon_Run(Unit*)	{};

  void		NameStatVals();

  void		CopyPNEPtrs(Network* net, Environment* env);
  // don't copy the environment, so user can select a different env for testing

  void	UpdateAfterEdit();
  void	Initialize();
  void 	Destroy()	{ CutLinks(); }
  void 	InitLinks();
  void 	CutLinks();
  void	Copy_(const ClosestEventStat& cp);
  COPY_FUNS(ClosestEventStat, Stat);
  TA_BASEFUNS(ClosestEventStat);
};

class CyclesToSettle : public Stat {
  // ##COMPUTE_IN_TrialProcess ##LOOP_STAT Records number of cycles it took to settle
public:
  SettleProcess* settle;	// #READ_ONLY #NO_SAVE settle process to record
  StatVal	cycles;		// number of cycles to settle

  virtual void	GetSettleProc();

  void		Layer_Run()	{ }; // don't do anything!!

  void		Network_Stat();
  void		InitStat();
  void		Init();
  bool		Crit();

  void		NameStatVals();

  void	UpdateAfterEdit();
  void 	Initialize();
  void 	Destroy()		{ CutLinks(); }
  void 	InitLinks();
  void	CutLinks();
  void	Copy_(const CyclesToSettle& cp);
  COPY_FUNS(CyclesToSettle, Stat);
  TA_BASEFUNS(CyclesToSettle);
};

class ScriptStat : public Stat {
  // Use this stat for custom script-based stats
public:
  StatVal_List	vals;		// put stat results in this group
  String_Array	s_args;		// string-valued arguments to pass to script

  void		InitStat();
  void		Init();
  bool		Crit();

  void		NameStatVals();

  void	UpdateAfterEdit();
  void	Initialize()		{ };
  void 	Destroy()		{ };
  void 	InitLinks();
  void	Copy_(const ScriptStat& cp);
  COPY_FUNS(ScriptStat, Stat);
  TA_BASEFUNS(ScriptStat);
};

class CompareStat : public Stat {
  // An aggregator-like stat that compares the output of two other stats
public:
  Stat*		stat_1;		// first comparison stat
  Stat*		stat_2;		// second comparison stat
  float_RArray::DistMetric cmp_type;	// comparison type to perform
  StatVal	cmp;		// comparison value
  float		dist_tol;	// tolerance value for distance functions
  bool		norm;		// whether to normalize values (distance, inner prod)
  SimpleMathSpec pre_proc_1;	// step 1 of pre-processing to apply before comparision
  SimpleMathSpec pre_proc_2;	// step 2 of pre-processing to apply before comparision
  SimpleMathSpec pre_proc_3;	// step 3 of pre-processing to apply before comparision
  float_RArray	svals_1;	// #HIDDEN #NO_SAVE stat_1 values 
  float_RArray	svals_2;	// #HIDDEN #NO_SAVE stat_2 values 

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Stat();
  void		Layer_Run()	{ };	// don't loop over network

  virtual void	GetStatVals(Stat* st, float_RArray& sv);
  // get stat vals into an array

  void 	Initialize();
  void	Destroy()		{ CutLinks(); }
  void	InitLinks();
  void 	CutLinks();
  void 	Copy_(const CompareStat& cp);
  COPY_FUNS(CompareStat, Stat);
  TA_BASEFUNS(CompareStat);
};

class ComputeStat : public Stat {
  // performs simple math computations on one or two stats, aggs the results according to net_agg
public:
  Stat*		stat_1;		// first comparison stat
  Stat*		stat_2;		// second comparison stat (optional)
  StatVal	cpt;		// aggregated computed value (if not COPY)
  SimpleMathSpec pre_proc_1;	// step 1 of pre-processing to apply to each stat before computing
  SimpleMathSpec pre_proc_2;	// step 2 of pre-processing to apply to each stat before computing
  SimpleMathSpec pre_proc_3;	// step 3 of pre-processing to apply to each stat before computing
  SimpleMathSpec compute_1;	// step 1 of computation: stat1 is val stat2 is arg
  SimpleMathSpec compute_2;	// step 2 of computation: stat1 is val stat2 is arg 
  SimpleMathSpec compute_3;	// step 3 of computation: stat1 is val stat2 is arg 
  float_RArray	svals_1;	// #HIDDEN #NO_SAVE stat_1 values 
  float_RArray	svals_2;	// #HIDDEN #NO_SAVE stat_2 values 

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Stat();
  void		Layer_Run()	{ };	// don't loop over network

  virtual void	GetStatVals(Stat* st, float_RArray& sv);
  // get stat vals into an array

  void 	Initialize();
  void	Destroy()		{ CutLinks(); }
  void	InitLinks();
  void 	CutLinks();
  void 	Copy_(const ComputeStat& cp);
  COPY_FUNS(ComputeStat, Stat);
  TA_BASEFUNS(ComputeStat);
};

class CopyToEnvStat : public Stat {
  // an aggregator-like stat that copies stat values from source stat to data environment
public:
  Stat*		stat;		// stat to get copy vals from
  Environment* 	data_env;	// environment to store data in
  SimpleMathSpec pre_proc_1;	// step 1 of pre-processing to apply before storage
  SimpleMathSpec pre_proc_2;	// step 2 of pre-processing to apply before storage
  SimpleMathSpec pre_proc_3;	// step 3 of pre-processing to apply before storage
  float_RArray	svals;		// #HIDDEN #NO_SAVE stat values 

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Stat();
  void		Layer_Run()	{ };	// don't loop over network

  virtual void	InitEnv();	// #BUTTON initalize environment to hold data
  virtual void	GetStatVals(Stat* st, float_RArray& sv);
  // get stat vals into an array

  void 	Initialize();
  void	Destroy()		{ CutLinks(); }
  void	InitLinks();
  void 	CutLinks();
  void 	Copy_(const CopyToEnvStat& cp);
  COPY_FUNS(CopyToEnvStat, Stat);
  TA_BASEFUNS(CopyToEnvStat);
};

class EpochCounterStat : public Stat {
  // ##FINAL_STAT gets the epoch counter from the network
public:
  StatVal	epoch;	

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Stat();
  void		Layer_Run()	{ };	// don't loop over network

  void		NameStatVals();

  void	Initialize();
  void 	Destroy()		{ };
  void	Copy_(const EpochCounterStat& cp);
  COPY_FUNS(EpochCounterStat, Stat);
  TA_BASEFUNS(EpochCounterStat);
};

class ProcCounterStat : public Stat {
  // ##FINAL_STAT gets the process counters from a different process hierarchy
public:
  SchedProcess*	proc;		// process to get counters from
  StatVal_List	counters;	// group of counters of data

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Stat();
  void		Layer_Run()	{ };	// don't loop over network

  virtual bool	NeedsUpdated();

  void		NameStatVals();

  void	UpdateAfterEdit();
  void	Initialize();
  void 	Destroy()		{ CutLinks(); }
  void 	InitLinks();
  void	CutLinks();
  void	Copy_(const ProcCounterStat& cp);
  COPY_FUNS(ProcCounterStat, Stat);
  TA_BASEFUNS(ProcCounterStat);
};

class MaxActTrgStat : public Stat {
  // ##COMPUTE_IN_TrialProcess 0-1 error statistic, 1 if unit with max act has trg = 1
public:
  StatVal	mxat;			// max activation = target 0-1 err value

  void		RecvCon_Run(Unit*)	{ }; // don't do these!
  void		SendCon_Run(Unit*)	{ };

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Init();
  void		Layer_Run();		// only compute on TARG layers
  void 		Layer_Stat(Layer* lay);
  void		Unit_Run(Layer*)	{ }; // don't loop over units

  void		NameStatVals();

  void	Initialize();
  void 	Destroy()		{ };
  SIMPLE_COPY(MaxActTrgStat);
  COPY_FUNS(MaxActTrgStat, Stat);
  TA_BASEFUNS(MaxActTrgStat);
};

class UnitActRFStat : public Stat {
  /* unit activity receptive-field stat: gets rf of units in layer by act
     weighted avg over rf_layer acts, stores into enviro! */
public:
  Layer_MGroup	rf_layers;	// #LINK_GROUP layers to compute rf over
  Environment*	data_env;	// environment to store data into
  float_RArray	avg_norms;	// average normalizers (sum of unit activity vals)

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Stat();
  void		Layer_Run()	{ };	// don't loop over network

  virtual void	InitRFVals();	// #BUTTON initalize rf aggregation values
  
  virtual bool	CheckRFLayersInNet(); // make sure rf layers are all in network.

  void	UpdateAfterEdit();
  void 	Initialize();
  void	Destroy()		{ CutLinks(); }
  void	InitLinks();
  void 	CutLinks();
  void 	Copy_(const UnitActRFStat& cp);
  COPY_FUNS(UnitActRFStat, Stat);
  TA_BASEFUNS(UnitActRFStat);
};

class UnitEventRFStat : public Stat {
  // stores one event per unit, with each pattern element representing firing for each trial
public:
  EpochProcess* epoch_proc;	// #READ_ONLY #NO_SAVE epoch process for event count
  Environment*	data_env;	// environment to store data into

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Stat();
  void		Layer_Run()	{ };	// don't loop over network

  virtual void	GetEpochProc();
  virtual void	InitRFVals();	// #BUTTON initalize rf aggregation values
  
  void	UpdateAfterEdit();
  void 	Initialize();
  void	Destroy()		{ CutLinks(); }
  void	InitLinks();
  void 	CutLinks();
  void 	Copy_(const UnitEventRFStat& cp);
  COPY_FUNS(UnitEventRFStat, Stat);
  TA_BASEFUNS(UnitEventRFStat);
};

class UniquePatStat : public Stat {
  // An aggregator-like stat that counts the number of unique patterns in given stat
public:
  Stat*		pat_stat;	// stat that is the source of patterns
  Environment*	data_env;	// environment to store patterns into during computation
  float_RArray::DistMetric cmp_type;	// comparison type to perform to determine uniqueness
  StatVal	unq;		// number of unique patterns
  float		dist_tol;	// bitwise tolerance value for distance functions
  bool		norm;		// whether to normalize values (distance, inner prod)
  float		uniq_tol;	// overall distance tolerance to determine if unique or not
  SimpleMathSpec pre_proc_1;	// step 1 of pre-processing to apply before comparision
  SimpleMathSpec pre_proc_2;	// step 2 of pre-processing to apply before comparision
  SimpleMathSpec pre_proc_3;	// step 3 of pre-processing to apply before comparision
  float_RArray	svals;		// #HIDDEN #NO_SAVE stat values 

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Stat();
  void		Layer_Run()	{ };	// don't loop over network

  virtual void	GetStatVals(Stat* st, float_RArray& sv);
  // get stat vals into an array

  void 	Initialize();
  void	Destroy()		{ CutLinks(); }
  void	InitLinks();
  void 	CutLinks();
  void 	Copy_(const UniquePatStat& cp);
  COPY_FUNS(UniquePatStat, Stat);
  TA_BASEFUNS(UniquePatStat);
};

#endif	/* procs_extra_h */
